#!/usr/bin/env python

import json
import os
import re
import socket
import subprocess
import sys
import yaml
import base64
import grp
import pwd
import shutil
import os.path
import glob

###############################################################################
# Global variables
###############################################################################
default_apache2_config_dir = "/etc/apache2/conf.d"
default_apache2_config = "%s/apache2.conf" % default_apache2_config_dir
default_apache2_service_config_dir = "/var/run/apache2"
test_apache2_vhost = "/etc/apache2/sites-available/testvhost"
hook_name = os.path.basename(sys.argv[0])

juju_warning_header = """#
#    "             "
#  mmm   m   m   mmm   m   m
#    #   #   #     #   #   #
#    #   #   #     #   #   #
#    #   "mm"#     #   "mm"#
#    #             #
#  ""            ""
# This file is managed by Juju. Do not make local changes.
#"""


###############################################################################
# Supporting functions
###############################################################################


#------------------------------------------------------------------------------
# juju_log:  Convenience wrapper around juju-log
#------------------------------------------------------------------------------
def juju_log(msg="MARK"):
    subprocess.call(['juju-log', str(msg)])

#------------------------------------------------------------------------------
# open_port:  Convenience function to open a port in juju to
#             expose a service
#------------------------------------------------------------------------------
def open_port(port=None, protocol="TCP"):
    if port is None:
        return(None)
    return(subprocess.call(['open-port', "%d/%s"
                            % (int(port), protocol)]))


#------------------------------------------------------------------------------
# close_port:  Convenience function to close a port in juju to
#              unexpose a service
#------------------------------------------------------------------------------
def close_port(port=None, protocol="TCP"):
    if port is None:
        return(None)
    return(subprocess.call(['close-port', "%d/%s"
                            % (int(port), protocol)]))


#------------------------------------------------------------------------------
# config_get:  Returns a dictionary containing all of the config information
#              Optional parameter: scope
#              scope: limits the scope of the returned configuration to the
#                     desired config item.
#------------------------------------------------------------------------------
def config_get(scope=None):
    try:
        config_cmd_line = ['config-get']
        if scope is not None:
            config_cmd_line.append(scope)
        config_cmd_line.append('--format=json')
        config_data = json.loads(subprocess.check_output(config_cmd_line))
        if not config_data["servername"]:
            config_data["servername"] = run(
                    ["unit-get", "public-address"]).rstrip()

    except Exception, e:
        juju_log(e)
        config_data = None
    finally:
        return(config_data)


#------------------------------------------------------------------------------
# relation_get:  Returns a dictionary containing the relation information
#                Optional parameters: scope, relation_id
#                scope:        limits the scope of the returned data to the
#                              desired item.
#                unit_name:    limits the data ( and optionally the scope )
#                              to the specified unit
#                relation_id:  specify relation id for out of context usage.
#------------------------------------------------------------------------------
def relation_get(scope=None, unit_name=None, relation_id=None):
    try:
        relation_cmd_line = ['relation-get', '--format=json']
        if relation_id is not None:
            relation_cmd_line.extend(('-r', relation_id))
        if scope is not None:
            relation_cmd_line.append(scope)
        if unit_name is not None:
            relation_cmd_line.append('-')
            relation_cmd_line.append(unit_name)
        juju_log('Calling: %s' % relation_cmd_line)
        relation_data = json.loads(subprocess.check_output(relation_cmd_line))
    except Exception:
        relation_data = None
    finally:
        return(relation_data)


def relation_ids_get(relation_name=None):
    try:
        relation_cmd_line = ['relation-ids', '--format=json']
        if relation_name is not None:
            relation_cmd_line.append(relation_name)
        juju_log('Calling: %s' % relation_cmd_line)
        relation_ids = json.loads(subprocess.check_output(relation_cmd_line))
    except Exception:
        relation_ids = None
    finally:
        return(relation_ids)


def relation_list_get(relation_id=None):
    try:
        relation_cmd_line = ['relation-list', '--format=json']
        if relation_id is not None:
            relation_cmd_line.extend(('-r', relation_id))
        juju_log('Calling: %s' % relation_cmd_line)
        relations = json.loads(subprocess.check_output(relation_cmd_line))
    except Exception:
        relations = None
    finally:
        return(relations)


def all_relation_data_get(relation_name=None):
    relation_ids = relation_ids_get(relation_name)
    if relation_ids is None:
        return()
    try:
        all_relation_data = {}
        for rid in relation_ids:
            units = relation_list_get(relation_id=rid)
            for unit in units:
                all_relation_data[unit.replace('/', '-')] = relation_get(relation_id=rid, unit_name=unit)
    except Exception:
        all_relation_data = None
    finally:
        return(all_relation_data)


#------------------------------------------------------------------------------
# apt_get_install( package ):  Installs a package
#------------------------------------------------------------------------------
def apt_get_install(packages=None):
    if packages is None:
        return(False)
    cmd_line = ['apt-get', '-y', 'install', '-qq']
    cmd_line.append(packages)
    return(subprocess.call(cmd_line))


#------------------------------------------------------------------------------
# apt_get_purge( package ):  Purges a package
#------------------------------------------------------------------------------
def apt_get_purge(packages=None):
    if packages is None:
        return(False)
    cmd_line = ['apt-get', '-y', 'purge', '-qq']
    cmd_line.append(packages)
    return(subprocess.call(cmd_line))


#------------------------------------------------------------------------------
# service_apache2:  Convenience function to start/stop/restart/reload
#                   the apache2 service
#------------------------------------------------------------------------------
def service_apache2(action=None):
    if action is None:
        return(None)
    elif action == "check":
        retVal = subprocess.call(['/usr/sbin/apache2ctl', 'configtest'])
        if retVal == 1:
            return(False)
        elif retVal == 0:
            return(True)
        else:
            return(False)
    else:
        retVal = subprocess.call(['service', 'apache2', action])
        if retVal == 0:
            return(True)
        else:
            return(False)


def run(command, *args, **kwargs):
    try:
        output = subprocess.check_output(command, *args, **kwargs)
        return output
    except Exception, e:
        print str(e)
        raise


def enable_module(module=None):
    if module is None:
        return(True)
    if os.path.exists("/etc/apache2/mods-enabled/%s.load" % (module)):
        juju_log("Module already loaded: %s" % module)
        return(True)
    if not os.path.exists("/etc/apache2/mods-available/%s.load" % (module)):
        retVal = apt_get_install("libapache2-mod-%s" % (module))
        if retVal != 0:
            juju_log("Installing module %s failed" % module)
            return(False)
    retVal = subprocess.call(['/usr/sbin/a2enmod', module])
    if retVal != 0:
        return(False)
    if service_apache2("check"):
        service_apache2("reload")
        return(True)


def disable_module(module=None):
    if module is None:
        return(True)
    if not os.path.exists("/etc/apache2/mods-enabled/%s.load" % (module)):
        juju_log("Module already disabled: %s" % module)
        return(True)
    retVal = subprocess.call(['/usr/sbin/a2dismod', module])
    if retVal != 0:
        return(False)
    if service_apache2("check"):
        service_apache2("reload")
        return(True)


###############################################################################
# Hook functions
###############################################################################
def install_hook():
    for f in glob.glob('exec.d/*/charm-pre-install'):
        if os.path.isfile(f) and os.access(f, os.X_OK):
            subprocess.check_call(['sh', '-c', f])
    if not os.path.exists(default_apache2_service_config_dir):
        os.mkdir(default_apache2_service_config_dir, 0600)
    apt_get_install("python-jinja2")
    return (apt_get_install("apache2"))


def dump_data(data2dump, log_prefix):
    log_file = '/tmp/pprint-%s.log' % (log_prefix)
    if data2dump is not None:
        logFile = open(log_file, 'w')
        import pprint
        pprint.pprint(data2dump, logFile)
        logFile.close()


def get_reverseproxy_data(relation='reverseproxy'):
    relation_data = all_relation_data_get(relation_name=relation)
    reverseproxy_data = {}
    if relation_data is None or len(relation_data) == 0:
        return reverseproxy_data
    for unit_name in relation_data.keys():
        if 'port' not in relation_data[unit_name]:
            return reverseproxy_data

        # unit_name: <service-name>-<unit_number>
        # jinja2 templates require python-type variables, remove all characters
        # that do not comply
        unit_type = re.sub(r'(.*)-[0-9]*', r'\1', unit_name)
        unit_type = re.sub('[^a-zA-Z0-9_]*', '', unit_type)
        juju_log('unit_type: %s' % unit_type)

        host = relation_data[unit_name]['private-address']
        for config_setting in relation_data[unit_name].keys():
            config_key = '%s_%s' % (unit_type, config_setting)
            reverseproxy_data[config_key] = relation_data[unit_name][config_setting]
            reverseproxy_data[unit_type] = '%s:%s' % (host, relation_data[unit_name]['port'])
        if 'all_services' in relation_data[unit_name]:
            service_data = yaml.load(relation_data[unit_name]['all_services'])
            for service_item in service_data:
                service_name = service_item['service_name']
                service_port = service_item['service_port']
                service_key = '%s_%s' % (unit_type, service_name)
                service_key = re.sub('[^a-zA-Z0-9_]*', '', service_key)
                reverseproxy_data[service_key] = '%s:%s' % (host, service_port)
    relation_data.update(reverseproxy_data)
    return relation_data


def update_balancers():
    relation_data = all_relation_data_get(relation_name='balancer')
    config_data = config_get()
    if relation_data is None or len(relation_data) == 0:
        juju_log('No relation data exiting')
        return
    unit_dict = {}
    for unit_name in relation_data.keys():
        if 'port' not in relation_data[unit_name]:
            return
        unit_type = re.sub(r'(.*)-[0-9]*', r'\1', unit_name)
        host = relation_data[unit_name]['private-address']
        port = relation_data[unit_name]['port']
        if unit_type in unit_dict:
            current_units = unit_dict[unit_type]
            current_units.append('%s:%s' % (host, port))
            unit_dict[unit_type] = current_units
        else:
            unit_dict[unit_type] = ['%s:%s' % (host, port)]
    from jinja2 import Environment, FileSystemLoader
    template_env = Environment(loader=FileSystemLoader(os.path.join(os.environ['CHARM_DIR'], 'data')))
    for balancer_name in unit_dict.keys():
        balancer_host_file = '%s/%s.balancer' % (default_apache2_config_dir,
                                                 balancer_name)
        templ_vars = {
            'balancer_name': balancer_name,
            'balancer_addresses': unit_dict[balancer_name],
            'lb_balancer_timeout': config_data['lb_balancer_timeout'],
        }
        template = template_env.get_template(
            'balancer.template').render(templ_vars)
        juju_log("Writing file: %s with data: %s" % (balancer_host_file, templ_vars))
        with open(balancer_host_file, 'w') as balancer_config:
            balancer_config.write(str(template))


def update_nrpe_checks():
    config_data = config_get()
    if 'nagios_check_http_params' not in config_data or len(config_data['nagios_check_http_params']) == 0:
        juju_log("No nrpe configuration, skipping")
        return
    try:
        nagios_uid = pwd.getpwnam('nagios').pw_uid
        nagios_gid = grp.getgrnam('nagios').gr_gid
    except:
        juju_log("Nagios user not setup, exiting")
        return

    unit_name = os.environ['JUJU_UNIT_NAME'].replace('/', '-')
    nrpe_check_file = '/etc/nagios/nrpe.d/check_vhost.cfg'
    nagios_hostname = "%s-%s" % (config_data['nagios_context'], unit_name)
    nagios_logdir = '/var/log/nagios'
    nagios_exportdir = '/var/lib/nagios/export'
    nrpe_service_file = '/var/lib/nagios/export/service__%s_check_vhost.cfg' %\
        (nagios_hostname)
    if not os.path.exists(nagios_logdir):
        os.mkdir(nagios_logdir)
        os.chown(nagios_logdir, nagios_uid, nagios_gid)
    if not os.path.exists(nagios_exportdir):
        juju_log('Exiting as %s is not accessible' % nagios_exportdir)
        return
    for f in os.listdir(nagios_exportdir):
        if re.search('.*check_vhost.cfg', f):
            os.remove(os.path.join(nagios_exportdir, f))
    from jinja2 import Environment, FileSystemLoader
    template_env = Environment(loader=FileSystemLoader(os.path.join(os.environ['CHARM_DIR'], 'data')))
    templ_vars = {
        'nagios_hostname': nagios_hostname,
        'nagios_servicegroup': config_data['nagios_context'],
    }
    template = \
        template_env.get_template('nrpe_service.template').render(templ_vars)
    with open(nrpe_service_file, 'w') as nrpe_service_config:
        nrpe_service_config.write(str(template))
    with open(nrpe_check_file, 'w') as nrpe_check_config:
        nrpe_check_config.write("# check Apache vhosts\n")
        nrpe_check_config.write(
            "command[check_vhost]=/usr/lib/nagios/plugins/check_http %s\n" %
            (config_data['nagios_check_http_params']))
    if os.path.isfile('/etc/init.d/nagios-nrpe-server'):
        subprocess.call(['service', 'nagios-nrpe-server', 'reload'])


def create_mpm_workerfile():
    config_data = config_get()
    mpm_workerfile = '/etc/apache2/conf.d/000mpm-worker'
    from jinja2 import Environment, FileSystemLoader
    template_env = Environment(loader=FileSystemLoader(os.path.join(os.environ['CHARM_DIR'], 'data')))
    templ_vars = {
        'mpm_type': config_data['mpm_type'],
        'mpm_startservers': config_data['mpm_startservers'],
        'mpm_minsparethreads': config_data['mpm_minsparethreads'],
        'mpm_maxsparethreads': config_data['mpm_maxsparethreads'],
        'mpm_threadlimit': config_data['mpm_threadlimit'],
        'mpm_threadsperchild': config_data['mpm_threadsperchild'],
        'mpm_serverlimit': config_data['mpm_serverlimit'],
        'mpm_maxclients': config_data['mpm_maxclients'],
        'mpm_maxrequestsperchild': config_data['mpm_maxrequestsperchild'],
    }
    template = \
        template_env.get_template('mpm_worker.template').render(templ_vars)
    with open(mpm_workerfile, 'w') as mpm_config:
        mpm_config.write(str(template))


def create_security():
    config_data = config_get()
    securityfile = '/etc/apache2/conf.d/security'
    from jinja2 import Environment, FileSystemLoader
    template_env = Environment(loader=FileSystemLoader(os.path.join(os.environ['CHARM_DIR'], 'data')))
    templ_vars = {
        'juju_warning_header': juju_warning_header,
        'server_tokens': config_data['server_tokens'],
        'server_signature': config_data['server_signature'],
        'trace_enabled': config_data['trace_enabled'],
    }
    template = \
        template_env.get_template('security.template').render(templ_vars)
    with open(securityfile, 'w') as security_config:
        security_config.write(str(template))


def ship_logrotate_conf():
    config_data = config_get()
    logrotate_file = '/etc/logrotate.d/apache2'
    from jinja2 import Environment, FileSystemLoader
    template_env = Environment(loader=FileSystemLoader(os.path.join(os.environ['CHARM_DIR'], 'data')))
    templ_vars = {
        'juju_warning_header': juju_warning_header,
        'logrotate_rotate': config_data['logrotate_rotate'],
        'logrotate_count': config_data['logrotate_count'],
        'logrotate_dateext': config_data['logrotate_dateext'],
    }
    template = template_env.get_template('logrotate.conf.template').render(templ_vars)
    with open(logrotate_file, 'w') as logrotate_conf:
        logrotate_conf.write(str(template))


def config_changed():
    relationship_data = {}
    config_data = config_get()
    relationship_data.update(get_reverseproxy_data(relation='reverseproxy'))
    relationship_data.update(get_reverseproxy_data(relation='website-cache'))
    update_balancers()
    if config_data['enable_modules']:
        module_list = config_data['enable_modules'].split()
        for module in module_list:
            enable_module(module)

    if config_data['disable_modules']:
        module_list = config_data['disable_modules'].split()
        for module in module_list:
            disable_module(module)

    if config_data['mpm_type'] == 'worker':
        apt_get_install('apache2-mpm-worker')
        apt_get_purge('apache2-mpm-prefork')
    elif config_data['mpm_type'] == 'prefork':
        apt_get_install('apache2-mpm-prefork')
        apt_get_purge('apache2-mpm-worker')
    create_mpm_workerfile()
    create_security()
    ports = {'http': 80, 'https': 443}
    for proto in ports.keys():
        template_var = 'vhost_%s_template' % (proto)
        template_data = dict(config_data.items() + relationship_data.items())
        close_port(ports[proto])
        if template_var in config_data:
            vhost_name = '%s_%s' % (config_data['servername'], proto)
            vhost_file = '/etc/apache2/sites-available/%s' % (vhost_name)
            from jinja2 import Template
            template = Template(
                str(base64.b64decode(config_data[template_var])))
            juju_log("Writing file: %s with data: %s" % (vhost_file, template_data))
            with open(vhost_file, 'w') as vhost:
                vhost.write(str(template.render(template_data)))
            open_port(ports[proto])
        subprocess.call(['/usr/sbin/a2ensite', vhost_name])

    cert_file = None
    if config_data['ssl_certlocation']:
        source = os.path.join(
            os.environ['CHARM_DIR'], 'data', config_data['ssl_certlocation'])
        cert_file = '/etc/ssl/certs/%s' % \
            (config_data['ssl_certlocation'].rpartition('/')[2])
        if os.path.exists(source):
            shutil.copy(source, cert_file)
        else:
            juju_log("Certificate not found, ignoring: %s" % source)

    chain_file = None
    if config_data['ssl_chainlocation']:
        chain_file = '/etc/ssl/certs/%s' % \
            (config_data['ssl_chainlocation'].rpartition('/')[2])
        shutil.copy(os.path.join(os.environ['CHARM_DIR'], 'data',
                                 config_data['ssl_chainlocation']), chain_file)

    key_file = None
    if config_data['ssl_keylocation']:
        source = os.path.join(
            os.environ['CHARM_DIR'], 'data', config_data['ssl_keylocation'])
        key_file = '/etc/ssl/private/%s' % \
            (config_data['ssl_keylocation'].rpartition('/')[2])
        if os.path.exists(source):
            shutil.copy(source, key_file)
            os.chmod(key_file, 0440)
            os.chown(key_file,
                pwd.getpwnam('root').pw_uid, grp.getgrnam('ssl-cert').gr_gid)
        else:
            juju_log("Key file not found, ignoring: %s" % source)

    if config_data['ssl_cert'] and cert_file is not None:
        if config_data['ssl_cert'] == "SELFSIGNED" and key_file is not None:
            config_data['ssl_key'] = ""
            gen_cert = os.path.join(os.environ['CHARM_DIR'], 'scripts',
                    'gen-selfsigned-cert')
            run([gen_cert, key_file, cert_file])
        else:
            juju_log("Writing cert from config ssl_cert: %s" % cert_file)
            with open(cert_file, 'w') as f:
                f.write(str(base64.b64decode(config_data['ssl_cert'])))

    if config_data['ssl_key'] and key_file is not None:
        juju_log("Writing key from config ssl_key: %s" % key_file)
        with open(key_file, 'w') as f:
            f.write(str(base64.b64decode(config_data['ssl_key'])))

    # Disable the default website because we don't want people to see the
    # "It works!" page on production services and remove the
    # conf.d/other-vhosts-access-log conf.
    if os.path.exists("/etc/apache2/sites-enabled/000-default"):
        run(["/usr/sbin/a2dissite", "000-default"])
    if os.path.exists("/etc/apache2/conf.d/other-vhosts-access-log"):
        os.unlink("/etc/apache2/conf.d/other-vhosts-access-log")

    if service_apache2("check"):
        if config_data["config_change_command"] in ["reload", "restart"]:
            service_apache2(config_data["config_change_command"])

    update_nrpe_checks()
    ship_logrotate_conf()


def start_hook():
    if service_apache2("status"):
        return(service_apache2("restart"))
    else:
        return(service_apache2("start"))


def stop_hook():
    if service_apache2("status"):
        return(service_apache2("stop"))


def reverseproxy_interface(hook_name=None):
    if hook_name is None:
        return(None)
    if hook_name == "changed":
        config_changed()


def website_interface(hook_name=None):
    if hook_name is None:
        return(None)
    my_host = socket.getfqdn(socket.gethostname())
    if my_host == "localhost":
        my_host = socket.gethostname()
    default_port = 80
    subprocess.call(['relation-set', 'port=%d' % default_port,
                     'hostname=%s' % my_host])


###############################################################################
# Main section
###############################################################################
if hook_name == "install":
    install_hook()
elif hook_name == "config-changed" or hook_name == "upgrade-charm":
    config_changed()
elif hook_name == "start":
    start_hook()
elif hook_name == "stop":
    stop_hook()
elif hook_name == "reverseproxy-relation-broken":
    config_changed()
elif hook_name == "reverseproxy-relation-changed":
    config_changed()
elif hook_name == "reverseproxy-relation-joined":
    config_changed()
elif hook_name == "balancer-relation-broken":
    config_changed()
elif hook_name == "balancer-relation-changed":
    config_changed()
elif hook_name == "balancer-relation-joined":
    config_changed()
elif hook_name == "website-cache-relation-broken":
    config_changed()
elif hook_name == "website-cache-relation-changed":
    config_changed()
elif hook_name == "website-cache-relation-joined":
    config_changed()
elif hook_name == "website-relation-joined":
    website_interface("joined")
elif hook_name == "nrpe-external-master-relation-changed":
    update_nrpe_checks()
else:
    print "Unknown hook"
    sys.exit(1)
